#version 330 core

in vec2 TexCoords;
out vec4 fragColor;

uniform sampler2D sceneColor;
uniform sampler2D sceneDepth;
uniform sampler2D shadowCubemap;
uniform float farPlane;
uniform vec3 cameraPosition;
uniform vec3 lightPosition;
uniform mat4 inverseViewProjection;
uniform mat4 lightSpaceMatrix;

const float lightIntensity = 300.0; // Light intensity (Phi)
const float density = 0.1; // Density of the participating medium
const float decay = 1; // Decay factor
const float tau = 0.1; // Extinction coefficient
const int maxSteps = 100; // Maximum number of raymarching steps

/*
float visibilitySampling(vec3 rayPosition)
{
    vec3 lightToFrag = rayPosition - lightPosition;
    float distance   = length(lightToFrag);
    float storedDepth = texture(shadowCubemap, normalize(lightToFrag)).r;
    float currentDepth = distance / farPlane;
    if (currentDepth > storedDepth + 0.0005)
        return 0.0;
    return 1.0;
}
*/

float visbilitySampling2D(vec3 rayPosition) {
    vec4 fragPosLightSpace = lightSpaceMatrix * vec4(rayPosition, 1.0);
    vec3 projCoords = fragPosLightSpace.xyz / fragPosLightSpace.w;
    projCoords = projCoords * 0.5 + 0.5;

    float shadowDepth  = texture(shadowCubemap, projCoords.xy).r;
    float currentDepth = projCoords.z;
    if (currentDepth > shadowDepth + 0.0005)
        return 0.0; // occluded
    return 1.0;
}

void main() {
    float fragDepth = texture(sceneDepth, TexCoords).r;
    vec2 ndc = TexCoords * 2.0 - 1.0; 
    float clipZ = fragDepth * 2.0 - 1.0;  
    vec4 clipPos = vec4(ndc, clipZ, 1.0);
    vec4 fragWorldPosition = inverseViewProjection * clipPos;
    fragWorldPosition /= fragWorldPosition.w;

    //Debugging meaningful depth values
    //float linearDepth =  (2.0 * 0.01 * farPlane) / (farPlane + 0.01 - (fragDepth  * 2.0 - 1.0) * (farPlane - 0.01));
    //float depthVis = linearDepth / farPlane;

    vec3 rayDir = normalize(cameraPosition - fragWorldPosition.xyz);
    float stepSize = length(cameraPosition - fragWorldPosition.xyz) / maxSteps;
    vec3 rayPosition = fragWorldPosition.xyz;
    float accumulatedLight = 0.0;
    float distanceTraveled = 0.0;

    for (int i = 0; i < maxSteps; ++i) {
        rayPosition += rayDir * stepSize;
        distanceTraveled += stepSize;

        float visibility = visbilitySampling2D(rayPosition);
        if (visibility < 1.0) continue;

        float distanceToLight = length(rayPosition - lightPosition);
        float attenuation = lightIntensity / (distanceToLight * distanceToLight);
        float scattering = exp(-tau * distanceToLight) * exp(-tau * distanceTraveled);
        accumulatedLight += visibility * attenuation * scattering * stepSize;
    }

    vec3 sceneColor = texture(sceneColor, TexCoords).rgb;
    vec3 godRays = sceneColor + accumulatedLight * density;
    fragColor = vec4(godRays, 1.0);

}
