#version 430

struct DirectionalLight {
    vec3 direction;
    vec3 intensity;
};

struct PointLight {
    vec3 position;
    vec3 intensity;
    vec3 attenuationFactors;
};

uniform float ambientStrength;

struct Material {
    sampler2D diffuse;
    sampler2D specular;
    float shininess;
};

in vec3 worldNormal;
in vec3 worldPosition;
in vec2 faceUv;
in vec4 FragPosLightSpace;

uniform samplerCube depthMap;
uniform vec3 viewPosition;
uniform vec3 lightPosition;
uniform Material material;

uniform float far_plane;
uniform DirectionalLight directionalLights[6];
uniform uint directionalLightCount;

uniform PointLight pointLights[6];
uniform uint pointLightCount;


out vec4 fragColor;

float shadowCalculation(vec3 fragPos) {

    float shadow  = 0.0;
    float bias = max(0.05 * (1.0 - dot(worldNormal, fragPos)), 0.09);
    float samples = 4.0;
    float offset  = 0.1;
    vec3 fragToLight = fragPos - lightPosition;
    for(float x = -offset; x < offset; x += offset / (samples * 0.5))
    {
        for(float y = -offset; y < offset; y += offset / (samples * 0.5))
        {
            for(float z = -offset; z < offset; z += offset / (samples * 0.5))
            {
                float closestDepth = texture(depthMap, fragToLight + vec3(x, y, z)).r;
                closestDepth *= far_plane;   // undo mapping [0;1]
                float currentDepth = length(fragToLight);
                if(currentDepth - bias > closestDepth)
                shadow += 1.0;
            }
        }
    }
    return shadow /= (samples * samples * samples);
}
vec3 phongDirectionalIllumination(
    vec3 lightIntensity,
    vec3 lightDirection,
    float lightAttenuation,
    vec3 worldPosition,
    vec3 worldNormal
) {

    worldNormal = normalize(worldNormal);
    lightDirection = normalize(lightDirection);

    vec3 ambientComponent = ambientStrength * texture(material.diffuse, faceUv).rgb;

    float diff = max(dot(worldNormal, -lightDirection), 0.0);
    vec3 diffuseComponent = lightAttenuation * diff * texture(material.diffuse, faceUv).rgb * lightIntensity;

    vec3 viewDirection = normalize(viewPosition - worldPosition);
    vec3 reflectDirection = normalize(reflect(lightDirection, worldNormal));
    float spec = pow(max(dot(viewDirection, reflectDirection), 0.0), material.shininess);
    vec3 specularComponent = lightAttenuation * spec * texture(material.specular, faceUv).rgb * lightIntensity;

    float shadow = shadowCalculation(FragPosLightSpace.xyz);
    return ambientComponent + (1 - shadow) * (diffuseComponent + specularComponent);
}

vec3 phongIllumination(vec3 worldPosition, vec3 worldNormal) {
    vec3 summedLight = vec3(0);

    for (int i = 0; i != min(directionalLightCount, 6); i++) {
        summedLight += phongDirectionalIllumination(
            directionalLights[i].intensity,
            directionalLights[i].direction,
            1.0,
            worldPosition,
            worldNormal
        );
    }

    for (int i = 0; i != min(pointLightCount, 6); i++) {
        PointLight light = pointLights[i];
        vec3 direction = worldPosition - light.position;
        float dist = length(direction);
        float attenuation = 1.0 / (light.attenuationFactors.x + light.attenuationFactors.y * dist + light.attenuationFactors.z * dist * dist);

        summedLight += phongDirectionalIllumination(
            light.intensity,
            direction,
            attenuation,
            worldPosition,
            worldNormal
        );
    }

    return summedLight;
}

void main() {

    fragColor = vec4(phongIllumination(worldPosition, worldNormal), 1.0);

    vec3 fragToLight = FragPosLightSpace.xyz - lightPosition;
    float closestDepth = texture(depthMap, fragToLight).r;
    closestDepth *= far_plane;

   //fragColor = vec4(vec3(closestDepth / far_plane), 1.0);
}
