#version 430

struct DirectionalLight {
    vec3 direction;
    vec3 intensity;
};

struct PointLight {
    vec3 position;
    vec3 intensity;
    vec3 attenuationFactors;
};

in vec3 worldNormal;
in vec3 worldPosition;
in vec2 faceUv;
in vec4 FragPosLightSpace;

uniform sampler2D shadowMap;
uniform vec3 colorVec;
uniform vec3 viewPosition;
uniform vec3 lightPosition;

uniform float ambientStrength;
uniform float diffuseStrength;
uniform float specularStrength;
uniform float specularAlpha;

uniform DirectionalLight directionalLights[6];
uniform uint directionalLightCount;

uniform PointLight pointLights[6];
uniform uint pointLightCount;


out vec4 fragColor;

float shadowCalculation(vec4 fragPosLightSpace) {
    // transform to normalized device coords
    vec3 projCoords = fragPosLightSpace.xyz / fragPosLightSpace.z;

    //transform coordinates from [-1,1] to [0,1]
    projCoords = (projCoords * 0.5) + 0.5;

    float closestDepth = texture(shadowMap, projCoords.xy).r;
    float currentDepth = projCoords.z;

    float shadow = currentDepth > closestDepth ? 1.0 : 0.0;
    return shadow;
}
vec3 phongDirectionalIllumination(
    vec3 lightIntensity,
    vec3 lightDirection,
    float lightAttenuation,
    vec3 worldPosition,
    vec3 worldNormal
) {

    worldNormal= normalize( worldNormal );
    lightDirection= normalize( lightDirection );

    vec3 ambientComponent= ambientStrength * vec3(1.0);

    vec3 diffuseComponent= lightAttenuation* diffuseStrength* max(dot(worldNormal, -lightDirection), 0.0)* lightIntensity;

    vec3 viewDirection= normalize(viewPosition - worldPosition);
    vec3 reflectDirection= normalize(reflect(lightDirection, worldNormal));
    vec3 specularComponent= lightAttenuation* specularStrength * pow(max(dot(viewDirection, reflectDirection), 0.0), specularAlpha) * lightIntensity;

    float shadow = shadowCalculation(FragPosLightSpace);
    return ambientComponent+ (1 - shadow) * (diffuseComponent + specularComponent);
}

vec3 phongIllumination( vec3 worldPosition, vec3 worldNormal ) {
    vec3 summedLight= vec3(0);

    for(int i= 0; i!= min(directionalLightCount, 6); i++) {
        summedLight+= phongDirectionalIllumination(
            directionalLights[i].intensity,
            directionalLights[i].direction,
            1.0,
            worldPosition,
            worldNormal
        );
    }

    for(int i= 0; i!= min(pointLightCount, 6); i++) {
        PointLight light= pointLights[i];
        vec3 direction= worldPosition - light.position;
        float dist= length(direction);
        float attenuation= 1.0 / (light.attenuationFactors.x + light.attenuationFactors.y * dist + light.attenuationFactors.z * dist * dist);

        summedLight+= phongDirectionalIllumination(
            light.intensity,
            direction,
            attenuation,
            worldPosition,
            worldNormal
        );
    }

    return summedLight;
}

void main() {

    fragColor= vec4(colorVec * phongIllumination( worldPosition, worldNormal ), 1.0);
}
